/*
 * This is jack-keyboard 1.6, a virtual keyboard for JACK MIDI. For questions
 * and comments, contact <trasz@pin.if.uz.zgora.pl>. This code is public
 * domain, you can do with it whatever you want.
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include <assert.h>
#include <string.h>
#include <sysexits.h>
#include <jack/jack.h>
#include <jack/midiport.h>
#include <jack/ringbuffer.h>
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>

struct Note {
	int		currently_playing;	/* 1 if note is being played. */
	int		sustain;

	int		x;	/* Distance between the left edge of key and
				 * left edge of the window, in pixels. */
	int		w;	/* Width of the key, in pixels. */
	int		h;	/* Height of the key, in pixels. */
	int		white;	/* 1 if key is white; 0 otherwise. */
};

#define NNOTES 127

volatile struct Note notes[NNOTES];

/* Table containing note values for the keys. */
int		note_for_key[256];

#define VELOCITY_HIGH		100
#define VELOCITY_NORMAL		64
#define VELOCITY_LOW		32

int		velocity = VELOCITY_NORMAL;

#define OUTPUT_PORT_NAME	"midi_out"
#define INPUT_PORT_NAME		"midi_in"
#define PROGRAM_NAME		"jack-keyboard"
#define PROGRAM_VERSION		"1.6"

jack_port_t    *output_port;
jack_port_t    *input_port;
int		entered_number = -1;
int		space_depressed = 0;
int		maybe_stop_sustained_notes = 0;
int		note_being_pressed_using_mouse = -1;
int		connect_at_startup = 1;
int		allow_connecting_to_jack_keyboard = 0;
int		time_offsets_are_zero = 0;
jack_client_t  *jack_client = NULL;

#define NOTE_ON			0x90
#define NOTE_OFF		0x80
#define PROGRAM_CHANGE		0xC0
#define CONTROLLER		0xB0
#define RESET			0xFF
#define HOLD_PEDAL		64
#define ALL_SOUND_OFF		120
#define ALL_CONTROLLERS_OFF	121
#define ALL_NOTES_OFF		123
#define BANK_SELECT_MSB		0
#define BANK_SELECT_LSB		32

GtkWidget      *window = NULL;

struct MidiMessage {
	jack_nframes_t	time;
	int		len;	/* Length of MIDI message, in bytes. */
	unsigned char	data[3];
};

#define RINGBUFFER_SIZE		1024*sizeof(struct MidiMessage)

/* Will emit a warning if time between jack callbacks is longer than this. */
#define MAX_TIME_BETWEEN_CALLBACKS	0.1

/* Will emit a warning if execution of jack callback takes longer than this. */
#define MAX_PROCESSING_TIME	0.01

jack_ringbuffer_t *ringbuffer;

/* Number of currently used program. */
int		program = 0;

/* Number of currently selected bank. */
int		bank = 0;

/* Number of currently selected channel (0..15). */
int		channel = 0;

void draw_note(int key);
void queue_message(int b0, int b1, int b2);

double 
get_time(void)
{
	double		seconds;
	int		ret;
	struct timeval	tv;

	ret = gettimeofday(&tv, NULL);
	if (ret) {
		perror("gettimeofday");
		exit(EX_OSERR);
	}
	seconds = tv.tv_sec + tv.tv_usec / 1000000.0;

	return seconds;
}

gboolean
process_received_message_async(gpointer evp)
{
	int		i;
	struct MidiMessage *ev = (struct MidiMessage *)evp;

	if (ev->data[0] == RESET || (ev->data[0] == CONTROLLER &&
		(ev->data[1] == ALL_NOTES_OFF || ev->data[1] == ALL_SOUND_OFF))) {

		for (i = 0; i < NNOTES; i++) {
			if (notes[i].currently_playing) {
				notes[i].currently_playing = 0;
				draw_note(i);
			}
		}
	}

	if (ev->data[0] == NOTE_ON) {
		int note = ev->data[1];
		notes[note].currently_playing = 1;
		draw_note(note);
	}

	if (ev->data[0] == NOTE_OFF) {
		int note = ev->data[1];
		notes[note].currently_playing = 0;
		draw_note(note);
	}

	if (ev->len == 1)
		queue_message(ev->data[0], -1, -1);
	else if (ev->len == 2)
		queue_message(ev->data[0], ev->data[1], -1);
	else
		queue_message(ev->data[0], ev->data[1], ev->data[2]);

	return FALSE;
}

struct MidiMessage *
midi_message_from_midi_event(jack_midi_event_t event)
{
	struct MidiMessage *ev = malloc(sizeof(ev));

	if (ev == NULL) {
		perror("malloc");
		return NULL;
	}

	if (event.size == 1) {
		ev->len = 1;
		ev->data[0] = event.buffer[0];

	} else if (event.size == 2) {
		ev->len = 2;
		ev->data[0] = event.buffer[0];
		ev->data[1] = event.buffer[1];

	} else {
		ev->len = 3;
		ev->data[0] = event.buffer[0];
		ev->data[1] = event.buffer[1];
		ev->data[2] = event.buffer[2];
	}

	ev->time = event.time;

	return ev;
}

int 
process_callback(jack_nframes_t nframes, void *arg)
{
	static double	time_last_called = -1;
	double		time_before_processing;
	double		time_after_processing;
	jack_nframes_t	last_frame_time;
	void           *port_buffer;
	struct MidiMessage ev;
	int		read, events, t, i;
	unsigned char  *buffer;
	jack_midi_event_t event;

	time_before_processing = get_time();

	if (time_last_called < 0)
		time_last_called = time_before_processing;

	if (time_before_processing - time_last_called > MAX_TIME_BETWEEN_CALLBACKS) {
		fprintf(stderr, "Had to wait %fs for JACK callback; possible note loss.\n",
			time_before_processing - time_last_called);
	}

	port_buffer = jack_port_get_buffer(input_port, nframes);

	events = jack_midi_get_event_count(port_buffer, nframes);

	for (i = 0; i < events; i++) {
		read = jack_midi_event_get(&event, port_buffer, i, nframes);
		if (read) {
			fprintf(stderr, "jack_midi_event_get failed, RECEIVED NOTE LOST.\n");
			continue;
		}

		struct MidiMessage *ev = midi_message_from_midi_event(event);
		if (ev == NULL) {
			fprintf(stderr, "midi_message_from_midi_event failed, RECEIVED NOTE LOST.\n");
			continue;
		}

		g_idle_add(process_received_message_async, ev);
	}

	last_frame_time = jack_last_frame_time(jack_client);

	port_buffer = jack_port_get_buffer(output_port, nframes);

	jack_midi_clear_buffer(port_buffer, nframes);

	while (jack_ringbuffer_read_space(ringbuffer)) {
		read = jack_ringbuffer_peek(ringbuffer, (char *)&ev, sizeof(ev));

		if (read != sizeof(ev)) {
			fprintf(stderr, "Short read from the ringbuffer, possible note loss.\n");
			continue;
		}

		t = ev.time + nframes - last_frame_time;

		/* If computed time is too much into the future, we'll need
		   to send it later. */
		if (t >= (int)nframes)
			break;

		/* If computed time is < 0, we missed a cycle because of xrun. */
		if (t < 0)
			t = 0;

		if (time_offsets_are_zero)
			t = 0;

		jack_ringbuffer_read_advance(ringbuffer, sizeof(ev));

		buffer = jack_midi_event_reserve(port_buffer, t, ev.len, nframes);

		if (buffer == NULL) {
			fprintf(stderr, "jack_midi_event_reserve failed, NOTE LOST.\n");
			continue;
		}

		if (ev.len > 2)
			buffer[2] = ev.data[2];

		if (ev.len > 1)
			buffer[1] = ev.data[1];

		buffer[0] = ev.data[0];
	}

	time_after_processing = get_time();

	if (time_after_processing - time_before_processing > MAX_PROCESSING_TIME) {
		fprintf(stderr, "Processing took %fs; scheduling problem?\n",
			time_after_processing - time_before_processing);
	}

	time_last_called = time_after_processing;

	return 0;
}

void 
queue_message(int b0, int b1, int b2)
{
	int		written;
	struct MidiMessage ev;

	if (b1 == -1) {
		ev.len = 1;
		ev.data[0] = b0;

	} else if (b2 == -1) {
		ev.len = 2;
		ev.data[0] = b0;
		ev.data[1] = b1;

	} else {
		ev.len = 3;
		ev.data[0] = b0;
		ev.data[1] = b1;
		ev.data[2] = b2;
	}

	ev.time = jack_frame_time(jack_client);

	written = jack_ringbuffer_write(ringbuffer, (char *)&ev, sizeof(ev));

	if (written != sizeof(ev))
		fprintf(stderr, "Not enough space in the ringbuffer, NOTE LOST.\n");
}

gboolean 
draw_window_title(gpointer notused)
{
	int		i, off = 0;
	char		title[256];
	const char    **connected_ports;

	if (window == NULL)
		return FALSE;

	connected_ports = jack_port_get_connections(output_port);

	off += snprintf(title, sizeof(title) - off, "%s: bank %d, program %d",
			jack_get_client_name(jack_client), bank, program);

	if (connected_ports == NULL) {
		off += snprintf(title + off, sizeof(title) - off, ", NOT CONNECTED");

	} else {
		off += snprintf(title + off, sizeof(title) - off, ", connected to ");

		for (i = 0; connected_ports[i] != NULL; i++) {
			off += snprintf(title + off, sizeof(title) - off, "%s%s",
				    i == 0 ? "" : ", ", connected_ports[i]);
		}

		off += snprintf(title + off, sizeof(title) - off, ", channel %d", channel + 1);
	}

	if (connected_ports != NULL)
		free(connected_ports);

	gtk_window_set_title(GTK_WINDOW(window), title);

	return FALSE;
}

void
update_window_title(void)
{
	g_idle_add(draw_window_title, NULL);
}

int 
graph_order_callback(void *dummy)
{
	if (window != NULL)
		update_window_title();

	return 0;
}

/* Connects to the specified input port, disconnecting already connected ports. */
int connect_to_input_port(const char *port)
{
        int ret;

	if (!strcmp(port, jack_port_name(input_port)))
		return -1;

	if (!allow_connecting_to_jack_keyboard) {
		if (!strncmp(port, jack_port_name(input_port), strlen(PROGRAM_NAME)))
			return -1;
	}

        ret = jack_port_disconnect(jack_client, output_port);

        if (ret) {
                fprintf(stderr, "Cannot disconnect MIDI port.\n");

                return -2;
        }

        ret = jack_connect(jack_client, jack_port_name(output_port), port);

        if (ret) {
                fprintf(stderr, "Cannot connect to %s.\n", port);

                return -3;

        } else {
                fprintf(stderr, "Connected to %s.\n", port);
                update_window_title();

                return 0;
        }
}

void
connect_to_another_input_port(int up_not_down)
{
	const char    **available_midi_ports;
	const char    **connected_ports;
	const char     *current = NULL;
	int		i, max, current_index;

	available_midi_ports = jack_get_ports(jack_client, NULL, JACK_DEFAULT_MIDI_TYPE, JackPortIsInput);

	/* There will be at least one listening MIDI port - the one we create. */
	assert(available_midi_ports);

	/*
	 * max is the highest possible index into available_midi_ports[],
	 * i.e. number of elements - 1.
	 */
	for (max = 0; available_midi_ports[max + 1] != NULL; max++);

	/* Only one input port - our own. */
	if (max == 0) {
		fprintf(stderr, "No listening JACK MIDI input ports found.  "
			"Run some softsynth and press Insert or Delete key.\n");

		return;
	}

	connected_ports = jack_port_get_connections(output_port);

	if (connected_ports)
		current = connected_ports[0];
	else
		current = available_midi_ports[0];

	for (i = 0; i <= max && strcmp(available_midi_ports[i], current); i++);

	current_index = i;

	assert(!strcmp(available_midi_ports[current_index], current));

	/* XXX: rewrite. */
	if (up_not_down) {
		for (i = current_index + 1; i <= max; i++) {
			assert(available_midi_ports[i] != NULL);

			if (connect_to_input_port(available_midi_ports[i]) == 0)
				goto out;
		}

		for (i = 0; i <= current_index; i++) {
			assert(available_midi_ports[i] != NULL);

			if (connect_to_input_port(available_midi_ports[i]) == 0)
				goto out;
		}

	} else {
		for (i = current_index - 1; i >= 0; i--) {
			assert(available_midi_ports[i] != NULL);

			if (connect_to_input_port(available_midi_ports[i]) == 0)
				goto out;
		}

		for (i = max; i >= current_index; i--) {
			assert(available_midi_ports[i] != NULL);

			if (connect_to_input_port(available_midi_ports[i]) == 0)
				goto out;
		}

	}

	if (!connected_ports)
		fprintf(stderr, "Cannot connect to any of the input ports.\n");

out:
	free(available_midi_ports);
	if (connected_ports != NULL)
		free(connected_ports);
}

void 
connect_to_next_input_port(void)
{
	connect_to_another_input_port(1);
}

void 
connect_to_prev_input_port(void)
{
	connect_to_another_input_port(0);
}

void 
init_jack(int argc, char *argv[])
{
	int err;

	jack_client = jack_client_open(PROGRAM_NAME, JackNoStartServer, NULL);

	if (jack_client == NULL) {
		fprintf(stderr, "Could not connect to the JACK server; run jackd first?\n");
		exit(EX_UNAVAILABLE);
	}

	err = jack_set_process_callback(jack_client, process_callback, 0);
	if (err) {
		fprintf(stderr, "Could not register JACK process callback.\n");
		exit(EX_UNAVAILABLE);
	}

	err = jack_set_graph_order_callback(jack_client, graph_order_callback, 0);
	if (err) {
		fprintf(stderr, "Could not register JACK graph order callback.\n");
		exit(EX_UNAVAILABLE);
	}

	output_port = jack_port_register(jack_client, OUTPUT_PORT_NAME, JACK_DEFAULT_MIDI_TYPE,
		JackPortIsOutput, 0);

	if (output_port == NULL) {
		fprintf(stderr, "Could not register JACK output port.\n");
		exit(EX_UNAVAILABLE);
	}

	input_port = jack_port_register(jack_client, INPUT_PORT_NAME, JACK_DEFAULT_MIDI_TYPE,
		JackPortIsInput, 0);

	if (input_port == NULL) {
		fprintf(stderr, "Could not register JACK input port.\n");
		exit(EX_UNAVAILABLE);
	}

	if (jack_activate(jack_client)) {
		fprintf(stderr, "Cannot activate client.\n");
		exit(EX_UNAVAILABLE);
	}

	if (connect_at_startup)
		connect_to_next_input_port();
}

void 
draw_key(GtkWidget *widget, int x, int y, int w, int h, int is_white)
{
	GdkColor	black = {0, 0, 0, 0};
	GdkColor	white = {0, 65535, 65535, 65535};

	GdkGC          *gc = widget->style->fg_gc[0];

	if (is_white)
		gdk_gc_set_rgb_fg_color(gc, &white);
	else
		gdk_gc_set_rgb_fg_color(gc, &black);

	gdk_draw_rectangle(widget->window, gc, TRUE, x, y, w, h);
	gdk_gc_set_rgb_fg_color(gc, &black);
	gdk_draw_rectangle(widget->window, gc, FALSE, x, y, w, h);
}

void 
draw_note(int note)
{
	int		white = notes[note].white;

	if (notes[note].currently_playing)
		white = !white;

	draw_key(window, notes[note].x, 0, notes[note].w, notes[note].h, white);

	/* We need to redraw black keys that partially obscure the white one. */
	if (note < NNOTES - 2 && !notes[note + 1].white)
		draw_note(note + 1);

	if (note >= 0 && !notes[note - 1].white)
		draw_note(note - 1);
}

void 
stop_unsustained_notes(void)
{
	int		i;

	for (i = 0; i < NNOTES; i++) {
		if (notes[i].currently_playing && !notes[i].sustain) {
			notes[i].currently_playing = 0;
			queue_message(NOTE_OFF + channel, i, velocity);
			draw_note(i);
		}
	}
}

void 
stop_sustained_notes(void)
{
	int		i;

	for (i = 0; i < NNOTES; i++) {
		if (notes[i].currently_playing && notes[i].sustain) {
			notes[i].currently_playing = 0;
			queue_message(NOTE_OFF + channel, i, velocity);
			draw_note(i);
		}
	}
}

int 
press_key(int key)
{
	assert(key >= 0);
	assert(key < NNOTES);

	maybe_stop_sustained_notes = 0;

	if (space_depressed)
		notes[key].sustain = 1;
	else
		notes[key].sustain = 0;

	if (notes[key].currently_playing)
		return 0;

	notes[key].currently_playing = 1;

	queue_message(NOTE_ON + channel, key, velocity);
	draw_note(key);

	return 1;
}

int 
release_key(int key)
{
	assert(key >= 0);
	assert(key < NNOTES);

	maybe_stop_sustained_notes = 0;

	if (space_depressed)
		notes[key].sustain = 1;
	else
		notes[key].sustain = 0;

	if (!notes[key].currently_playing)
		return 0;

	if (notes[key].sustain)
		return 0;

	notes[key].currently_playing = 0;

	queue_message(NOTE_OFF + channel, key, velocity);
	draw_note(key);

	return 1;
}

void 
panic(void)
{
	int		i;

	/*
	 * These two have to be sent first, in case we have no room in the
	 * ringbuffer for all these NOTE_OFF messages sent five lines below.
	 */
	queue_message(CONTROLLER + channel, ALL_NOTES_OFF, 0);
	queue_message(CONTROLLER + channel, ALL_SOUND_OFF, 0);

	for (i = 0; i < NNOTES; i++) {
		release_key(i);

		queue_message(NOTE_OFF + channel, i, 0);
		usleep(100);
	}

	queue_message(CONTROLLER + channel, HOLD_PEDAL, 0);
	queue_message(CONTROLLER + channel, ALL_CONTROLLERS_OFF, 0);
	queue_message(CONTROLLER + channel, ALL_NOTES_OFF, 0);
	queue_message(CONTROLLER + channel, ALL_SOUND_OFF, 0);
	queue_message(RESET, -1, -1);
}

unsigned char 
char_for_keyval(guint kv)
{
	gchar          *name;

	switch (kv) {
		case GDK_space:
			return ' ';
		case GDK_minus:
		case GDK_KP_Subtract:
			return '-';
		case GDK_plus:
		case GDK_KP_Add:
			return '+';
		case GDK_equal:
			return '=';
		case GDK_KP_Multiply:
			return '*';
		case GDK_KP_Divide:
			return '/';
		default:
			name = gdk_keyval_name(gdk_keyval_to_lower(kv));
			if (strlen(name) == 1)
				return name[0];

			return 0;
	}
}

void
add_digit(int digit)
{
	if (entered_number == -1)
		entered_number = 0;
	else
		entered_number *= 10;

	entered_number += digit;
}

gint 
keyboard_event_handler(GtkWidget *widget, GdkEventKey *event, gpointer func_data)
{
	/* 0 for octave #4 (as in C4) */
	static int	octave_shift = 0;
	int		i;
	unsigned char	key;
	guint		keyval;
	GdkKeymapKey	kk;

	/* We're not using event->keyval, because we need keyval with level set to 0.
	   E.g. if user holds Shift and presses '7', we want to get a '7', not '&'. */
	kk.keycode = event->hardware_keycode;
	kk.level = 0;
	kk.group = 0;

	keyval = gdk_keymap_lookup_key(NULL, &kk);

	key = char_for_keyval(keyval);

	window = widget;

	/*
	 * '+' key shifts octave up. '-' key shifts octave down.
	 */
	if (key == '+' || key == '=') {
		if (event->type == GDK_KEY_PRESS && octave_shift < 3) {
			stop_unsustained_notes();
			octave_shift++;
		}

		return TRUE;
	}

	if (key == '-') {
		if (event->type == GDK_KEY_PRESS && octave_shift > -5) {
			stop_unsustained_notes();
			octave_shift--;
		}

		return TRUE;
	}
	/*
	 * Shift increases velocity, i.e. holding it while pressing note key
	 * will make the sound louder. Ctrl decreases velocity.
	 */
	if (event->keyval == GDK_Shift_L || event->keyval == GDK_Shift_R) {
		if (event->type == GDK_KEY_PRESS)
			velocity = VELOCITY_HIGH;
		else
			velocity = VELOCITY_NORMAL;

		return TRUE;

	}

	if (event->keyval == GDK_Control_L || event->keyval == GDK_Control_R) {
		if (event->type == GDK_KEY_PRESS)
			velocity = VELOCITY_LOW;
		else
			velocity = VELOCITY_NORMAL;

		return TRUE;
	}

	/*
	 * User can enter a number from the keypad; after that, pressing
	 * '*'/'/', Page Up/Page Down etc will set program/bank/whatever
	 * to the number that was entered.
	 */
	/*
	 * XXX: This is silly.  Find a way to enter the number without
	 * all these contitional instructions.
	 */
	if (event->keyval == GDK_KP_0 || event->keyval == GDK_KP_Insert) {
		if (event->type == GDK_KEY_PRESS)
			add_digit(0);

		return TRUE;
	}

	if (event->keyval == GDK_KP_1 || event->keyval == GDK_KP_End) {
		if (event->type == GDK_KEY_PRESS)
			add_digit(1);

		return TRUE;
	}

	if (event->keyval == GDK_KP_2 || event->keyval == GDK_KP_Down) {
		if (event->type == GDK_KEY_PRESS)
			add_digit(2);

		return TRUE;
	}

	if (event->keyval == GDK_KP_3 || event->keyval == GDK_KP_Page_Down) {
		if (event->type == GDK_KEY_PRESS)
			add_digit(3);

		return TRUE;
	}

	if (event->keyval == GDK_KP_4 || event->keyval == GDK_KP_Left) {
		if (event->type == GDK_KEY_PRESS)
			add_digit(4);

		return TRUE;
	}

	if (event->keyval == GDK_KP_5 || event->keyval == GDK_KP_Begin) {
		if (event->type == GDK_KEY_PRESS)
			add_digit(5);

		return TRUE;
	}

	if (event->keyval == GDK_KP_6 || event->keyval == GDK_KP_Right) {
		if (event->type == GDK_KEY_PRESS)
			add_digit(6);

		return TRUE;
	}

	if (event->keyval == GDK_KP_7 || event->keyval == GDK_KP_Home) {
		if (event->type == GDK_KEY_PRESS)
			add_digit(7);

		return TRUE;
	}

	if (event->keyval == GDK_KP_8 || event->keyval == GDK_KP_Up) {
		if (event->type == GDK_KEY_PRESS)
			add_digit(8);

		return TRUE;
	}

	if (event->keyval == GDK_KP_9 || event->keyval == GDK_KP_Page_Up) {
		if (event->type == GDK_KEY_PRESS)
			add_digit(9);

		return TRUE;
	}

	/*
	 * '*' character increases program number. '/' character decreases
	 * it.
	 */
	if (key == '*') {
		if (event->type == GDK_KEY_PRESS) {
			if (entered_number > -1)
				program = entered_number;
			else
				program++;

			if (program < 0)
				program = 0;

			if (program > 127)
				program = 127;

			queue_message(PROGRAM_CHANGE + channel, program, -1);
			update_window_title();

			entered_number = -1;
		}

		return TRUE;
	}

	if (key == '/') {
		if (event->type == GDK_KEY_PRESS) {
			if (entered_number > -1)
				program = entered_number;
			else
				program--;

			if (program < 0)
				program = 0;

			if (program > 127)
				program = 127;

			queue_message(PROGRAM_CHANGE + channel, program, -1);
			update_window_title();

			entered_number = -1;
		}

		return TRUE;
	}

	if (event->keyval == GDK_Page_Up) {
		if (event->type == GDK_KEY_PRESS) {
			if (entered_number > -1)	
				bank = entered_number;
			else
				bank++;

			if (bank < 0)
				bank = 0;

			if (bank > 16383)
				bank = 16383;

			queue_message(CONTROLLER + channel, BANK_SELECT_LSB, bank % 128);
			queue_message(CONTROLLER + channel, BANK_SELECT_MSB, bank / 128);
			queue_message(PROGRAM_CHANGE + channel, program, -1);
			update_window_title();

			entered_number = -1;
		}

		return TRUE;
	}

	if (event->keyval == GDK_Page_Down) {
		if (event->type == GDK_KEY_PRESS) {
			if (entered_number > -1)
				bank = entered_number;
			else
				bank--;

			if (bank < 0)
				bank = 0;

			if (bank > 16383)
				bank = 16383;

			queue_message(CONTROLLER + channel, BANK_SELECT_LSB, bank % 128);
			queue_message(CONTROLLER + channel, BANK_SELECT_MSB, bank / 128);
			queue_message(PROGRAM_CHANGE + channel, program, -1);
			update_window_title();

			entered_number = -1;
		}

		return TRUE;
	}

	if (event->keyval == GDK_Home) {
		if (event->type == GDK_KEY_PRESS && channel < 15) {
			channel++;
			update_window_title();
		}

		return TRUE;
	}

	if (event->keyval == GDK_End) {
		if (event->type == GDK_KEY_PRESS && channel > 0) {
			channel--;
			update_window_title();
		}

		return TRUE;
	}

	if (event->keyval == GDK_Insert) {
		if (event->type == GDK_KEY_PRESS)
			connect_to_next_input_port();

		return TRUE;
	}

	if (event->keyval == GDK_Delete) {
		if (event->type == GDK_KEY_PRESS)
			connect_to_prev_input_port();

		return TRUE;
	}

	if (event->keyval == GDK_Escape) {
		if (event->type == GDK_KEY_PRESS)
			panic();

		return TRUE;
	}

	/*
	 * Spacebar works as a 'sustain' key.  Holding spacebar while
	 * releasing note will cause the note to continue. Pressing and
	 * releasing spacebar without pressing any note keys will make all
	 * the sustained notes end (it will send 'note off' midi messages for
	 * all the sustained notes).
	 */
	if (key == ' ') {
		if (event->type == GDK_KEY_PRESS) {
			if (!space_depressed) {
				space_depressed = 1;
				maybe_stop_sustained_notes = 1;
			}

		} else {
			if (maybe_stop_sustained_notes)
				stop_sustained_notes();

			space_depressed = 0;
		}

		return TRUE;
	}

	if (key == 0) {
		fprintf(stderr, "Unknown key '%s'; ignored.\n", gdk_keyval_name(event->keyval));
		return TRUE;
	}

	i = note_for_key[key];

	if (i < 0) {
		fprintf(stderr, "No binding for key '%c'; ignored.\n", key);
		return TRUE;
	}

	i += octave_shift * 12;

	assert(i >= 0);
	assert(i <= NNOTES);

	if (event->type == GDK_KEY_PRESS) {
		press_key(i);

	} else if (event->type == GDK_KEY_RELEASE) {
		release_key(i);
	}

	return TRUE;
}

int 
get_note_for_xy(int x, int y)
{
	int		width;
	int		height;
	int		note;

	gtk_window_get_size(GTK_WINDOW(window), &width, &height);

	if (y <= height / 2) {
		for (note = 0; note < NNOTES - 1; note++) {
			if (notes[note].white)
				continue;

			if (x >= notes[note].x && x <= notes[note].x + notes[note].w)
				return note;
		}
	}

	for (note = 0; note < NNOTES - 1; note++) {
		if (!notes[note].white)
			continue;

		if (x >= notes[note].x && x <= notes[note].x + notes[note].w)
			return note;
	}

	return -1;
}

gboolean 
mouse_button_event_handler(GtkWidget *widget, GdkEventButton *event, gpointer user_data)
{
	int		x = event->x;
	int		y = event->y;

	int		note = get_note_for_xy(x, y);

	if (event->type == GDK_BUTTON_PRESS) {
		/* This is possible when you make the window a little wider and then click
		   on the grey area. */
		if (note < 0) {
			return TRUE;
		}

		if (note_being_pressed_using_mouse >= 0)
			release_key(note_being_pressed_using_mouse);

		press_key(note);
		note_being_pressed_using_mouse = note;

	} else if (event->type == GDK_BUTTON_RELEASE) {
		if (note >= 0) {
			release_key(note);

		} else {
			if (note_being_pressed_using_mouse >= 0)
				release_key(note_being_pressed_using_mouse);
		}

		note_being_pressed_using_mouse = -1;

	}

	return TRUE;
}

gboolean 
mouse_motion_event_handler(GtkWidget *widget, GdkEventMotion *event, gpointer user_data)
{
	int		note;

	if ((event->state & GDK_BUTTON1_MASK) == 0)
		return TRUE;

	note = get_note_for_xy(event->x, event->y);

	if (note != note_being_pressed_using_mouse && note >= 0) {
		
		if (note_being_pressed_using_mouse >= 0)
			release_key(note_being_pressed_using_mouse);
		press_key(note);
		note_being_pressed_using_mouse = note;
	}

	return TRUE;
}

gboolean 
expose_event(GtkWidget *widget, GdkEventExpose *event, gpointer user_data)
{
	int		width;
	int		height;

	/*
	 * NNOTES-1 is the number of notes.  There are 7 white keys per 12
	 * notes in octave.
	 */
	int		number_of_white_keys = (NNOTES - 1) * (7.0 / 12.0);

	int		key_width;
	int		black_key_width;

	int		note;
	int		white_key = 0;
	int		white;

	gtk_window_get_size(GTK_WINDOW(widget), &width, &height);

	key_width = width / number_of_white_keys;
	black_key_width = key_width * 0.8;

	for (note = 0, white_key = 0; note < NNOTES - 2; note++) {
		int		note_in_octave = note % 12;

		if (notes[note].currently_playing) {
			white = 0;

		} else {
			white = 1;
		}

		if (note_in_octave == 1 || note_in_octave == 3 || note_in_octave == 6 ||
			note_in_octave == 8 || note_in_octave == 10) {

			/* This note is black key; draw them later. */
			continue;
		}

		/* This note is white key. */

		notes[note].x = white_key * key_width;
		notes[note].w = key_width;
		notes[note].h = height;
		notes[note].white = 1;

		draw_key(widget, notes[note].x, 0, notes[note].w, notes[note].h, white);

		white_key++;
	}

	/* Draw black keys over the white ones. */
	for (note = 0, white_key = 0; note < NNOTES - 2; note++) {
		int		note_in_octave = note % 12;

		if (notes[note].currently_playing) {
			white = 0;

		} else {
			white = 1;
		}

		if (note_in_octave == 1 || note_in_octave == 3 || note_in_octave == 6 ||
			note_in_octave == 8 || note_in_octave == 10) {

			/* This note is black key. */
			white = !white;

			notes[note].x = white_key * key_width - black_key_width / 2;
			notes[note].w = black_key_width;
			notes[note].h = height / 2;
			notes[note].white = 0;

			draw_key(widget, notes[note].x, 0, notes[note].w, notes[note].h, white);

			continue;
		}

		white_key++;
	}

	return TRUE;
}

void 
init_notes(void)
{
	int		i;

	for (i = 0; i <= 255; i++)
		note_for_key[i] = -1;

	/* Lower keyboard row - 'zxcvbnm'. */
	note_for_key['z'] = 60;	/* C4 */
	note_for_key['s'] = 61;
	note_for_key['x'] = 62;
	note_for_key['d'] = 63;
	note_for_key['c'] = 64;
	note_for_key['v'] = 65;
	note_for_key['g'] = 66;
	note_for_key['b'] = 67;
	note_for_key['h'] = 68;
	note_for_key['n'] = 69;
	note_for_key['j'] = 70;
	note_for_key['m'] = 71;

	/* Higher keyboard row, first octave - 'qwertyu'. */
	note_for_key['q'] = 72;
	note_for_key['2'] = 73;
	note_for_key['w'] = 74;
	note_for_key['3'] = 75;
	note_for_key['e'] = 76;
	note_for_key['r'] = 77;
	note_for_key['5'] = 78;
	note_for_key['t'] = 79;
	note_for_key['6'] = 80;
	note_for_key['y'] = 81;
	note_for_key['7'] = 82;
	note_for_key['u'] = 83;

	/* Higher keyboard row, the rest - 'iop'. */
	note_for_key['i'] = 84;
	note_for_key['9'] = 85;
	note_for_key['o'] = 86;
	note_for_key['0'] = 87;
	note_for_key['p'] = 88;
}

void
show_version(void)
{
	fprintf(stdout, "jack-keyboard %s\n", PROGRAM_VERSION);

	exit(EX_OK);
}

void
usage(void)
{
	fprintf(stderr, "usage: jack-keyboard [-Vnkt] [-c <channel>] [-b <bank> ] [-p <program>]\n");
	fprintf(stderr, "   where <channel> is MIDI channel to use for output, from 1 to 16,\n");
	fprintf(stderr, "   <bank> is MIDI bank to use, from 0 to 16383,\n");
	fprintf(stderr, "   and <program> is MIDI program to use, from 0 to 127\n");

	exit(EX_USAGE);
}

int 
main(int argc, char *argv[])
{
	int ch;

	while ((ch = getopt(argc, argv, "Vnktc:b:p:")) != -1) {
		switch (ch) {
			case 'V':
				show_version();
				break;

			case 'n':
				connect_at_startup=0;
				break;

			case 'k':
				allow_connecting_to_jack_keyboard=1;
				break;
				
			case 't':
				time_offsets_are_zero=1;
				break;
				
			case 'c':
				channel = atoi(optarg);

				if (channel < 1 || channel > 16) {
					fprintf(stderr, "Invalid MIDI channel number; "
						"valid values are 1-16.\n");

					exit(EX_USAGE);
				}

				channel--; /* 1..16 -> 0..15 */
				break;

			case 'b':
				bank = atoi(optarg);

				if (bank < 0 || bank > 16383) {
					fprintf(stderr, "Invalid MIDI bank number; "
						"valid values are 0-16383.\n");

					exit(EX_USAGE);
				}

				break;
	
			case 'p':
				program = atoi(optarg);

				if (program < 0 || program > 127) {
					fprintf(stderr, "Invalid MIDI program number; "
						"valid values are 0-127.\n");

					exit(EX_USAGE);
				}

				break;
				
			case '?':
			default:
				usage();
		}
	}

	argc -= optind;
	argv += optind;

	memset((void *)notes, 0, sizeof(struct Note) * NNOTES);

	ringbuffer = jack_ringbuffer_create(RINGBUFFER_SIZE);

	if (ringbuffer == NULL) {
		fprintf(stderr, "Cannot create ringbuffer.\n");
		exit(EX_SOFTWARE);
	}

	jack_ringbuffer_mlock(ringbuffer);

	init_jack(argc, argv);
	init_notes();

	g_thread_init(NULL);
	gtk_init(&argc, &argv);

	window = gtk_window_new(GTK_WINDOW_TOPLEVEL);

	update_window_title();

	gtk_widget_set_app_paintable(window, TRUE);
	gtk_window_set_default_size(GTK_WINDOW(window), 730, 70);
	gtk_widget_add_events(window, GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK | GDK_POINTER_MOTION_MASK);
	g_signal_connect(G_OBJECT(window), "destroy", G_CALLBACK(exit), NULL);
	g_signal_connect(G_OBJECT(window), "expose-event", G_CALLBACK(expose_event), NULL);
	g_signal_connect(G_OBJECT(window), "button-press-event", G_CALLBACK(mouse_button_event_handler), NULL);
	g_signal_connect(G_OBJECT(window), "button-release-event", G_CALLBACK(mouse_button_event_handler), NULL);
	g_signal_connect(G_OBJECT(window), "motion-notify-event", G_CALLBACK(mouse_motion_event_handler), NULL);
	g_signal_connect(G_OBJECT(window), "key-press-event", G_CALLBACK(keyboard_event_handler), NULL);
	g_signal_connect(G_OBJECT(window), "key-release-event", G_CALLBACK(keyboard_event_handler), NULL);
	gtk_widget_show(window);

	gtk_main();

	return 0;
}

